package net.sf.sido.io.xml;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import net.sf.sido.DataFactory;
import net.sf.sido.DataObject;
import net.sf.sido.DataType;
import net.sf.sido.Property;
import net.sf.sido.PropertyMetaInfos;
import net.sf.sido.io.AbstractDataObjectReader;

import org.dom4j.Attribute;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.QName;
import org.dom4j.io.SAXReader;

public class XMLDataObjectReader extends AbstractDataObjectReader<Element> {

	private final XMLDataObjectNamespace namespace;

	public XMLDataObjectReader() {
		this(new DefaultXMLDataObjectNamespace());
	}

	public XMLDataObjectReader(DefaultXMLDataObjectNamespace n) {
		namespace = n;
	}

	private int getAttInteger(Element e, QName name, boolean required, int defaultValue) {
		String value = getAttString(e, name, required, null);
		if (value != null) {
			return Integer.parseInt(value, 10);
		} else {
			return defaultValue;
		}
	}

	private String getAttString(Element e, QName name, boolean required, String defaultValue) {
		Attribute att = e.attribute(name);
		if (att != null) {
			return att.getStringValue();
		} else if (required) {
			throw new DOXmlAttributeNotFoundException(name, e.getName());
		} else {
			return defaultValue;
		}
	}

	@Override
	protected int getDOId(Element e) {
		return getAttInteger(e, XMLDOSchema.ATT_ID, false, 0);
	}

	@Override
	protected int getDOReference(Element e) {
		return getAttInteger(e, XMLDOSchema.ATT_REF, false, 0);
	}

	@Override
	protected String getDOType(Element e, boolean required) {
		return getAttString(e, XMLDOSchema.ATT_TYPE, required, null);
	}

	@Override
	protected Element getElement(Element e, Property<?> property) {
		QName pName = namespace.getQName(property);
		return e.element(pName);
	}

	@Override
	protected List<Element> getElements(Element e, Property<?> property) {
		QName pName = namespace.getQName(property);
		// Gets the element that holds the array
		Element eArray = e.element(pName);
		if (eArray == null) {
			return null;
		} else {
			// Data object property?
			boolean doProperty = property.getMetaInfo().getBoolean(PropertyMetaInfos.DO, false);
			// Element QName to use
			QName eName = doProperty ? XMLDOSchema.ELE_DATAOBJECT : XMLDOSchema.ELE_VALUE;
			// Gets the elements
			@SuppressWarnings("unchecked")
			List<Element> elements = eArray.elements(eName);
			// OK
			return elements;
		}
	}

	@Override
	protected String getText(Element eNode) {
		return eNode.getText();
	}

	@Override
	public DataObject read(InputStream in, DataFactory factory, DataType expectedType) throws IOException {
		try {
			// Reads the XML
			Document dom = new SAXReader().read(in);
			// Reads the DOM
			return read(dom.getRootElement(), factory, expectedType);
		} catch (DocumentException ex) {
			throw new IOException("Cannot parse XML", ex);
		}
	}

}
