package net.sf.sido.support;

import java.io.IOException;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import net.sf.sido.DOUndefinedPropertyException;
import net.sf.sido.DataFactory;
import net.sf.sido.DataObject;
import net.sf.sido.DataType;
import net.sf.sido.Property;
import net.sf.sido.PropertyAccessor;
import net.sf.sido.PropertyMetaInfos;
import net.sf.sido.array.Array;
import net.sf.sido.editor.DataEditor;
import net.sf.sido.io.DataObjectIO;
import net.sf.sido.io.json.JSONDataObjectWriter;

import org.apache.commons.lang.ObjectUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.joda.time.DateTime;
import org.joda.time.Duration;
import org.joda.time.Period;

import com.google.common.base.Function;

public abstract class AbstractDataObject implements DataObject, Serializable {

    private static final long serialVersionUID = 1;

	private final int id;
	private final DataType type;

    // Deserialization fields
    protected transient int t_id;
    protected transient String t_typename;

	public AbstractDataObject(int id, DataType type) {
		Validate.notNull(type, "Data object type is required");
		if (type.isAbstractType()) {
			throw new DOCannotInstantiateAbstractTypeException(type);
		}
		this.id = id;
		this.type = type;
	}

    private void writeObject(java.io.ObjectOutputStream out) throws IOException {
        out.writeUTF(type.getQualifiedName());
        out.writeInt(id);
    }

    private void readObject(java.io.ObjectInputStream in)
     throws IOException, ClassNotFoundException {
        t_typename = in.readUTF();
        t_id = in.readInt();
    }

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof DataObject) {
			DataObject o = (DataObject) obj;
			DataType oType = o.getSidoType();
			if (oType.equals(type)) {
				// Gets all properties
				Collection<Property<?>> properties = type.getProperties();
				for(Property<?> property : properties) {
					Object thisValue = getValue(property);
					Object thatValue = o.getValue(property);
					if (!ObjectUtils.equals(thisValue, thatValue)) {
						return false;
					}
				}
				// OK
				return true;
			} else {
				return false;
			}
		} else {
			return false;
		}
	}

	@Override
	public DataObject copy(DataObject from, boolean copy, String... properties) {
		DataType fromType = from.getSidoType();
		for(String name : properties) {
			// Gets the property definition
			Property<Object> p = fromType.getProperty(name);
			// Gets the property value
			Object value = from.getValue(p);
			if (copy) {
				// Array of DO
				if (p.isArray() && p.isDO()) {
					@SuppressWarnings("unchecked")
					Array<DataObject> array = (Array<DataObject>) value;
					// Copy of this array
					array = array.transform(new Function<DataObject, DataObject>() {

						@Override
						public DataObject apply(DataObject i) {
							return i.copy();
						}

					});
					// OK
					value = array;
				}
				// DO
				else if (p.isDO()) {
					value = ((DataObject) value).copy();
				}
			}
			// Sets the property here
			setValue(name, value);
		}
		// OK
		return this;
	}

	@Override
	public DataObject copy(DataObject from, String... properties) {
		return copy(from, false, properties);
	}

	@Override
	public Object get(String name) {
		return getObject(name);
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> Array<T> getArray(String property) {
		return (Array<T>) getObject(property);
	}

	public <T> DataObject add(String property, T... elements) {
		// Gets the array
		Array<T> array = getArray(property);
		// Completes the array
		if (array == null) {
			array = new Array<T>(elements);
		} else {
			array = array.add(elements);
		}
		// Resets the array
		setArray(property, array);
		// OK
		return this;
	}

	@Override
	public Array<DataObject> getDOArray(String property) {
		return getArray(property);
	}

	@Override
	public Boolean getBoolean(String property) {
		return (Boolean) getObject(property);
	}

	@Override
	public Byte getByte(String name) {
		return (Byte) getObject(name);
	}

	@Override
	public DataObject getDataObject(String property) {
		return (DataObject) getObject(property);
	}

	@Override
	public DateTime getDateTime(String name) {
		return (DateTime) getObject(name);
	}

	@Override
	public BigDecimal getDecimal(String property) {
		return (BigDecimal) getObject(property);
	}

	@Override
	public Double getDouble(String name) {
		return (Double) getObject(name);
	}

	@Override
	public Duration getDuration(String name) {
		return (Duration) getObject(name);
	}

	@Override
	public DataFactory getFactory() {
		return type.getSchema().getFactory();
	}

	@Override
	public Float getFloat(String name) {
		return (Float) getObject(name);
	}

	@Override
	public Integer getInteger(String property) {
		return (Integer) getObject(property);
	}

	@Override
	public Long getLong(String name) {
		return (Long) getObject(name);
	}

	@Override
	public Period getPeriod(String name) {
		return (Period) getObject(name);
	}

	@Override
	public Short getShort(String name) {
		return (Short) getObject(name);
	}

	@Override
	public int getSidoId() {
		return id;
	}

	@Override
	public String getString(String property) {
		return (String) getObject(property);
	}

	@Override
	public DataType getSidoType() {
		return type;
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> T getValue(Property<T> property) {
		String name = property.getName();
		return (T) getObject(name);
	}

	@Override
	public Object getValue(String propertyPath) {
		return getObject(propertyPath);
	}

	@Override
	public int hashCode() {
		return toString().hashCode();
	}

	@Override
	public void init() {
		init(false);
	}

	@Override
	public <T> DataObject setArray(String property, Array<T> value) {
		return setObject(property, value);
	}

	@Override
	public <T> DataObject setArray(String name, Collection<T> list) {
		Array<T> array = new Array<T>(list);
		return setArray(name, array);
	}

	@Override
	public DataObject setAsString(String name, String... values) {
		// Gets the property
		Property<?> p = type.getProperty(name);
		if (p == null) {
			throw new DOUndefinedPropertyException(this, name);
		}
		// Checks the property
		if (p.isDO()) {
			throw new IllegalArgumentException("setAsString cannot be called for DataObject property like " + name);
		}
		// Array property
		if (p.isArray()) {
			// List of values
			List<Object> list = new ArrayList<Object>();
			// Gets the element type
			Class<? super Object> elementType = p.getMetaInfo().getType(PropertyMetaInfos.ELEMENT_TYPE);
			// Gets the editor for the element type
			DataEditor<?> editor = getFactory().getEditor(elementType);
			// For each value
			for(String value : values) {
				// Converts the value
				Object t = editor.valueOf(value);
				// ... and adds it to the list
				list.add(t);
			}
			// OK, sets the value
			setArray(name, list);
		}
		// Simple property
		else {
			// Checks the number of values
			if (values == null || values.length == 0) {
				setValue(name, null);
			} else if (values.length > 1) {
				throw new IllegalArgumentException("Cannot call setAsString with several values for a simple property: " + name);
			} else {
				String value = values[0];
				// Gets the editor for the property type
				DataEditor<?> editor = getFactory().getEditor(p.getType());
				// Object value
				Object t = editor.valueOf(value);
				// Sets the value
				setValue(name, t);
			}
		}
		// OK
		return this;
	}

	@Override
	public DataObject setBoolean(String property, Boolean value) {
		return setObject(property, value);
	}

	@Override
	public DataObject setByte(String name, Byte value) {
		return setObject(name, value);
	}

	@Override
	public DataObject setDataObject(String property, DataObject value) {
		return setObject(property, value);
	}

	@Override
	public DataObject setDateTime(String name, DateTime value) {
		return setObject(name, value);
	}

	@Override
	public DataObject setDecimal(String property, BigDecimal value) {
		return setObject(property, value);
	}

	@Override
	public DataObject setDouble(String name, Double value) {
		return setObject(name, value);
	}

	@Override
	public DataObject setDuration(String name, Duration value) {
		return setObject(name, value);
	}

	@Override
	public DataObject setFloat(String name, Float value) {
		return setObject(name, value);
	}

	@Override
	public DataObject setInteger(String property, Integer value) {
		return setObject(property, value);
	}

	@Override
	public DataObject setLong(String name, Long value) {
		return setObject(name, value);
	}

	@Override
	public DataObject setPeriod(String name, Period value) {
		return setObject(name, value);
	}

	@Override
	public DataObject setShort(String name, Short value) {
		return setObject(name, value);
	}

	@Override
	public DataObject setString(String property, String value) {
		return setObject(property, value);
	}

	@Override
	public <T> DataObject setValue(Property<T> property, T value) {
		return setObject(property.getName(), value);
	}

	@Override
	public PropertyAccessor setValue(String propertyPath, Object value) {
		return setObject(propertyPath, value);
	}

	@Override
	public <E extends Enum<E>> DataObject setEnum(String name, E value) {
		// Converts to string
		String s = value != null ? value.toString() : null;
		// Sets the value
		return setString(name, s);
	}

	@Override
	public <E extends Enum<E>> E getEnum(String name, Class<E> enumClass) {
		// Gets the value as string
		String s = getString(name);
		// Conversion
		if (StringUtils.isBlank(s)) {
			return null;
		} else {
			return Enum.valueOf(enumClass, s);
		}
	}

	@Override
	public String toString() {
		try {
			return DataObjectIO.writeString(new JSONDataObjectWriter(), this);
		} catch (IOException e) {
			return e.getMessage();
		}
	}
}
